{% extends "_base.html" %}
{% block title %}{{ super() }} {% trans %}Processes{% endtrans %} {% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="./processes">{% trans %}Processes{% endtrans %}</a>
{% endblock %}
{% block body %}
    <section id="processes">
      <h2>{% trans %}Processes in this service{% endtrans %}</h2>
      <div class="row">
        <div class="col-sm-12 col-md-6">
          <table class="table table-striped table-bordered">
            <thead>
              <tr>
                <th>{% trans %}Name{% endtrans %}</th>
                <th>{% trans %}Description{% endtrans %}</th>
              </tr>
            </thead>
            <tbody>
              {% for p in data['processes'] %}
              <tr itemscope itemtype="https://schema.org/WebAPI">
                <td itemprop="name" data-label="name">
                  <meta itemprop="url" content="{{ config['server']['url'] }}/processes/{{ p['id'] }}" />
                  <a title="{{ p['title'] | striptags | truncate }}" href="{{ config['server']['url'] }}/processes/{{ p['id'] }}">{{ p['title'] | striptags | truncate }}</a>
                </td>
                <td itemprop="description" data-label="description">
                  {{ p['description'] | striptags | truncate }}
                </td>
              </tr>
              {% endfor %}
            </tbody>
          </table>
        </div>
      </div>
    </section>
{% endblock %}
