{% extends "_base.html" %}
{% block title %}{{ super() }} {% trans %}SpatioTemporal Asset Catalog{% endtrans %} {% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="./stac">{% trans %}SpatioTemporal Asset Catalog{% endtrans %}</a>
{% endblock %}
{% block body %}

    <section id="identification">
      <h4>{% trans %}STAC Version{% endtrans %}: {{ data['stac_version'] }}</h4>
    </section>

    <section id="collections">
      <h2>Collections</h2>
            <table class="table table-striped table-bordered">
              <thead>
              <tr>
                <th>{% trans %}Name{% endtrans %}</th>
                <th>{% trans %}Description{% endtrans %}</th>
              </tr>
              </thead>
              <tbody>
                {% for link in data['links'] %}
                {% if link['rel'] == 'child' and link['type'] == 'application/json' %}
                <tr>
                  <td data-label="name">
                    <a  title="{{ link['title'] | striptags | truncate }}"
                      href="{{ config['server']['url'] }}/stac/{{ link['title'] }}">
                      <span>{{ link['title'] | striptags | truncate }}</span></a>
                  </td>
                  <td data-label="description">
                    {{ link['description'] | striptags | truncate }}
                  </td>
                </tr>
                {% endif %}
                {% endfor %}
              </tbody>
            </table>
    </section>

{% endblock %}
